#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

//wrapper for .8xv appvar file
unsigned char _8xvWrapper[] =	{
				'*', '*', 'T', 'I', '8', '3', 'F', '*',
				0x1A, 0x0A, 0x00,
				'C','r','e','a','t','e','d',' ','b','y',' ','a','m','i','h',
				'a','r','t','.',0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
				//lower then upper bytes of SIZE+17 : offset|53
					0x00, 0x00, 
				//<outer data section> : offset|55
					0x0D, 0X00, 
					//lower then upper bytes of SIZE : offset|57
					0x00, 0x00, 
					0x15,
					//appvar name 
					0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
					0x00, 0x80,
					//lower then upper bytes of SIZE : offset|70
					0x00, 0x00, 
					//lower then upper bytes of SIZE : offset|72
					0x00, 0x00, 
					//<inner data section> : offset|74
					//..............
					//</inner data section>
				//</outer data section>
				//lower then upper bytes of checksum off : offset|76+SIZE
				0x00,0x00
			};

//calculate a checksum from some data
int genChecksum(int size, unsigned char* data) {
	int sum = 0;
	int i;
	for (i=55; i<74+size; i++)
		sum += data[i];

	while (sum >= 65536)
		sum -= 65536;

	return sum;
}

//gets the lower byte of a two-byte number
int lowerByte(int data) {
	return data%256;
}
//gets the upper byte of a two-byte number
int upperByte(int data) {
	return (data-(data%256))/256;
}

//generates .8xv appvar binary data from data
unsigned char* gen8xv(unsigned char* name, int size, unsigned char* data) {
	int charLen = size+sizeof(_8xvWrapper)/sizeof(_8xvWrapper[0]);
	int totalSize = 76+size;
	unsigned char* _8xv = malloc(sizeof(unsigned char)*totalSize);
	int i=0;
	for (i=0; i<74; i++)
		_8xv[i] = _8xvWrapper[i];

	_8xv[53] = lowerByte(size+17);
	_8xv[54] = upperByte(size+17);
	_8xv[57] = lowerByte(size);
	_8xv[58] = upperByte(size);
	_8xv[70] = lowerByte(size);
	_8xv[71] = upperByte(size);
	_8xv[72] = lowerByte(size-2);
	_8xv[73] = upperByte(size-2);
	for (i=0; i<sizeof(name)/sizeof(name[0]); i++) 
		_8xv[60+i] = name[i];
	for (i=0; i<size; i++)
		_8xv[i+74] = data[i];
	_8xv[size+74] = lowerByte(genChecksum(size,_8xv));
	_8xv[size+75] = upperByte(genChecksum(size,_8xv));
	return _8xv;
}

//create an appvar .8xv file from some data
void create8xv(unsigned char* fname, unsigned char* name, int dataSize, unsigned char* data) {
	int writeDataSize = dataSize+4;
	unsigned char writeData[writeDataSize];
	writeData[0] = 01;
	writeData[1] = 03;
	writeData[2] = 01;
	writeData[3] = 02;
	for (int i=4; i<dataSize; i++)
		writeData[i]=data[i-4];
	unsigned char* outData = gen8xv(name,writeDataSize,writeData);
	FILE *fp = fopen(fname,"wb");
	fwrite(outData,sizeof(unsigned char),76+writeDataSize,fp);
	fclose(fp);
	free(outData);
}

int usage() {
	printf("ppm8xv [PPM file name]\n");
	printf("\n");
	printf("This software will convert a 320x240 raw PPM image\n");
	printf("file into three AppVars, each containing one-third\n");
	printf("of the image at the full 65k color depth.\n");
	printf("\n");
}

void throwError(char* error) {
	printf("%s\nType `ppm8xv --help` for usage.\n",error);
	return;
}

//converts three-byte RGB to two-byte RGB
int RGB(double r, double g, double b) {

	r = floor(r/8);
	g = floor(g/4);
	b = floor(b/8);

	double rgb = r*pow(2,11) + g*pow(2,5) + b ;
	return (int)rgb;
}

//Alters a string so that it only contains
//capital letters and numbers. Lowercase 
//letters become capital and everything
//else is dropped. 
char* fixVarName(char* name, int nameLen) {
	if (nameLen > 7)
		nameLen = 7;

	int fixedLen = 0;
	for (int i=0; i<nameLen; i++)
		if ( 
			(name[i] >= '0' && name[i] <= '9') ||
			(name[i] >= 'A' && name[i] <= 'Z') ||
			(name[i] >= 'a' && name[i] <= 'z')
		   )
			fixedLen++;
	char* out = malloc(sizeof(char)*(fixedLen+1));
	out[fixedLen] = 0;
	int d = 0;
	for (int i=0; i<fixedLen; i++) {
		if ( 
			(name[i] >= '0' && name[i] <= '9') ||
			(name[i] >= 'A' && name[i] <= 'Z') ||
			(name[i] >= 'a' && name[i] <= 'z')
		   ) 
			if (name[i] >= 'a' && name[i] <= 'z')
				out[d++] = name[i]-32;
			else
				out[d++] = name[i];
	}
	return out;
}

//Adds a number (0-9) to the end of a string
char* addNumber(char* name,int number) {
	int len = strlen(name);
	char* out = malloc(sizeof(char)*(len+2));
	for (int i=0; i<len; i++)
		out[i] = name[i];
	out[len] = (char)(number+48);
	out[len+1] = 0;
	return out;
}

//Adds ".8xv" to the end of a string
char* add8xv(char* name) {
	int len = strlen(name);
	char* out = malloc(sizeof(char)*(len+5));
	for (int i=0; i<len; i++)
		out[i] = name[i];
	out[len] = '.';
	out[len+1] = '8';
	out[len+2] = 'x';
	out[len+3] = 'v';
	out[len+4] = 0;
	return out;
}

int main(int argc, char** args) {

	if (argc < 2) { 
		throwError("No file specified.");
		return 1;
	}

	if (!strcmp(args[1],"--help")) {
		usage();
		return 0;
	}

	FILE *file = fopen(args[1],"rb");
	if (file == NULL) {
		throwError("Specified file not found.");
		return 1;
	}

	int bareNameLen = 0;
	while (args[1][bareNameLen] != '.')
		bareNameLen++;
	
	char bareName[bareNameLen+1];
	bareName[bareNameLen] = 0;
	for (int i=0; i<bareNameLen; i++) 	
		bareName[i] = args[1][i];

	char* bareNameFixed = fixVarName(bareName,bareNameLen);
	char* varName1 = addNumber(bareNameFixed,1);
	char* varName2 = addNumber(bareNameFixed,2);
	char* varName3 = addNumber(bareNameFixed,3);
	free(bareNameFixed);

	char* fileBareName1 = addNumber(bareName,1);
	char* fileBareName2 = addNumber(bareName,2);
	char* fileBareName3 = addNumber(bareName,3);
	char* fileName1 = add8xv(fileBareName1);
	char* fileName2 = add8xv(fileBareName2);
	char* fileName3 = add8xv(fileBareName3);
	free(fileBareName1);
	free(fileBareName2);
	free(fileBareName3);

	fseek(file,0,SEEK_END);
	int size=ftell(file);
	rewind(file);

	unsigned char buffer[size];
	fread(buffer,size,1,file);
	
	int dataSectionAddr=0;
	int lineCount=0;
	int width=-1;
	int height=-1;

	if (buffer[0] != 'P' || buffer[1] != '6') {
		throwError("Unrecognized file format.");
		return 1;
	}

	while (lineCount<3) {
		if (buffer[dataSectionAddr] == (int)'#') {
			lineCount--;
		}

		if (lineCount == 1 && width==-1 && height==-1) {
			int sizeOfFirst = 0;
			int sizeOfSecond = 0;
			while (buffer[dataSectionAddr+sizeOfFirst] != (char)' ') 
				sizeOfFirst++;
			while (buffer[dataSectionAddr+sizeOfFirst+sizeOfSecond+1] != (char)'\n') 
				sizeOfSecond++;

			char first[sizeOfFirst];
			char second[sizeOfSecond];

			for (int i=0; i<sizeOfFirst; i++)
				first[i] = buffer[dataSectionAddr+i];
			for (int i=0; i<sizeOfSecond; i++)
				second[i] = buffer[dataSectionAddr+sizeOfFirst+1+i];

			width = atoi(first);
			height = atoi(second);

		}

		if (buffer[dataSectionAddr] == (int)'\n') 
			lineCount++;
		
		dataSectionAddr++;
	}

	if (width != 320 || height != 240) {
		throwError("Image is not the correct dimensions.");
		return 1;
	}

	int dataSize = 320*240*2;
	unsigned char data[dataSize];

	int d = 0;
	for (int i=dataSectionAddr; i<size; i+=3) {
		int rgb = RGB(buffer[i],buffer[i+1],buffer[i+2]);
		data[d++] = lowerByte(rgb);
		data[d++] = upperByte(rgb);
		
	}

	int sec1 = dataSize/3;
	int sec2 = 2*(dataSize/3);
	unsigned char subData1[dataSize/3];
	unsigned char subData2[dataSize/3];
	unsigned char subData3[dataSize/3];

	for (int i=0; i<sec1; i++)
		subData1[i] = data[i];
	for (int i=sec1; i<sec2; i++)
		subData2[i-sec1] = data[i];
	for (int i=sec2; i<dataSize; i++)
		subData3[i-sec2] = data[i];
	
	create8xv(fileName1,varName1,dataSize/3,subData1);
	create8xv(fileName2,varName2,dataSize/3,subData2);
	create8xv(fileName3,varName3,dataSize/3,subData3);

	free(fileName1);
	free(fileName2);
	free(fileName3);
	free(varName1);
	free(varName2);
	free(varName3);

}

